from __future__ import print_function
from __future__ import division

import os
import sys

import yaml
import civis
import click

from .utils import get_most_recent_run
from .defaults import LASTSCRIPTID_LOC


def _print_script_runs(scriptid):
    """Print the runs for a script.

    Parameters
    ----------
    scriptid : int
        The container script to print the runs for.
    """
    client = civis.APIClient(resources='all')

    try:
        # First get all of the runs.
        runs = client.scripts.list_containers_runs(scriptid)

        fmt = "%- 25s %- 25s %- 25s %- 25s %s"
        head = fmt % ("run id", "started", "finished", "status", "error msg")
        print(head)
        for run in runs:
            print(fmt % (run['id'],
                         run['started_at'] if run['started_at'] else '-',
                         run['finished_at'] if run['finished_at'] else '-',
                         run['state'],
                         run['error'] if run['error'] else '-'))
    except civis.base.CivisAPIError as e:
        print(
            "Could not get script %s runs: %s.%s: %s" % (
                scriptid, e.__module__, e.__class__.__name__, e),
            file=sys.stderr)
        sys.exit(-1)


def _print_script_logs(scriptid, runid=None):
    """Print the logs and information for a script.

    Parameters
    ----------
    scriptid : int
        The script ID to print out.
    runid : int or None, optional
        An optional run ID. If None, then the most recent
        run will be used.
    """
    client = civis.APIClient(resources='all')

    try:
        # get the run details
        runid = runid or get_most_recent_run(scriptid, client=client)

        # Get rest of script details.
        deets = client.scripts.get_containers(scriptid)

        # Make them into pure dicts for dumping to yaml.
        deets = {k: dict(v)
                 if (isinstance(v, civis.response.Response) or
                     isinstance(v, dict)) else v
                 for k, v in deets.items()}

        # Put the name and id at the top so they are easy to see
        print('name:', deets['name'])
        print('id:', deets['id'])
        del deets['name']
        del deets['id']

        # Make the docker command more grep-able.
        print('docker_command:')
        for line in deets['docker_command'].split('\n'):
            line = line.rstrip()
            print('  ' + line)
        del deets['docker_command']

        # now dump the rest.
        print(yaml.dump(deets, default_flow_style=False).strip())

        if runid is not None:
            # Get and print the logs for this run
            resp = client.scripts.list_containers_runs_logs(scriptid, runid)
            print("log file:")
            for r in resp[::-1]:
                print("  [%s] %s" % (r['created_at'], r['message']))
        else:
            print('log file: -')
    except civis.base.CivisAPIError as e:
        print(
            "Could not print script %s logs: %s.%s: %s" % (
                scriptid, e.__module__, e.__class__.__name__, e),
            file=sys.stderr)
        sys.exit(-1)


def _print_scripts(user_ids, state, hidden):
    """List scripts given a list of users.

    This function only lists scripts that are containers.

    Parameters
    ----------
    user_ids : list of ints
        List of user IDs to get jobs for.
        An empty list corresponds to all scripts
        visible to the user making the API call.
    state : str
        List scripts only in this state. Use None to
        get scripts in all states.
    hidden : bool
        If True, display hidden scripts in addition to non-hidden ones.
    """
    client = civis.APIClient(resources='all')

    # Get scripts.
    scripts = []

    def _append_scripts(_hidden):
        _scripts = client.scripts.list(
            limit=50,
            hidden=_hidden,
            status=state,
            order='updated_at',
            type='containers',
            author=','.join(str(u) for u in user_ids) if user_ids else None)
        for s in _scripts:
            scripts.append(s)

    _append_scripts(_hidden=False)
    if hidden:
        _append_scripts(_hidden=True)

    # Print them out.
    fmt_str = "%- 20s %- 20s %- 20s %- 25s %- 25s %- 25s %s"
    head = fmt_str % (
        "id",
        "author",
        "status",
        "created",
        "started",
        "finished",
        "name")
    print(head)

    for s in scripts:
        if user_ids and s['author']['id'] not in user_ids:
            continue

        if s['last_run'] is not None:
            s_tme = s['last_run']['startedAt'] or '-'
            f_tme = s['last_run']['finishedAt'] or '-'
        else:
            s_tme = '-'
            f_tme = '-'
        c_tme = s['created_at'] or '-'

        print(fmt_str % (s['id'],
                         s['author']['username'],
                         s['state'],
                         c_tme,
                         s_tme,
                         f_tme,
                         s['name']))


def _parse_user_ids(users, client):
    """parse a string of users separated by commas into IDS"""
    ids = []
    split_users = users.split(',')
    for u in split_users:
        u = u.strip()
        if len(u) > 0:
            uid = client.users.list(query=u)
            if uid:
                ids.append(uid[0]['id'])
    return ids


def _status(scriptid, mine, user, runs, runid, last, hidden, state):
    """Inspect a civis-compute script.

    If `scriptid` is given, print information about the script IDs
    according to the various other arguments. Otherwise, the status
    of the civis-compute queue is printed.

    Parameters
    ----------
    scriptid : int or None
        The container script ID to inspect.
    mine : bool
        If listing the status of the civis-compute queue, only list containers
        where the current user is the author.
    user : str or None
        A comma-separated list of user names. If given and the status of the
        civis-compute queue is being printed, only list containers by the given
        users.
    runs : bool
        If `scriptid` is given, print the runs for the script.
    runid : int or None
        If give and `runs` is False, print the script logs for a given run.
    last : bool
        If True, print the script logs for the last script submitted at the
        command line.
    hidden : bool
        If True and the status of the civis-compute queue is being printed,
        include hidden scripts in the outputs.
    state : str
        If the status of the civis-compute queue is being printed, then list
        only scripts with this state.
    """

    if runs and runid is not None:
        raise click.UsageError(
            "Only one of `--runs` or `--run-id` can be given!")

    if scriptid is not None and last:
        raise click.UsageError(
            "Only one of `--last` or `scriptid` can be given!")

    if last:
        try:
            with open(os.path.expandvars(os.path.expanduser(
                    LASTSCRIPTID_LOC)), 'r') as fp:
                scriptid = int(fp.read().strip())
        except:
            print('The most recently run script ID could not be found!',
                  file=sys.stderr)
            sys.exit(-1)

    if scriptid is None:
        client = civis.APIClient(resources='all')

        user_ids = []
        if mine:
            user_ids += [client.users.list_me()['id']]

        if user:
            user_ids += _parse_user_ids(user, client)

        _print_scripts(
            user_ids,
            state if state != 'all' else None,
            hidden)
    else:
        if runs:
            _print_script_runs(scriptid)
        else:
            _print_script_logs(scriptid, runid=runid)



from __future__ import print_function
from __future__ import division

from six import StringIO

import pytest

from ..parsing import (
    _parse_options_from_file,
    _join_container_params,
    _cleanup_required_resources,
    _reformat_files_key,
    _process_options)


@pytest.mark.parametrize(
    'line',
    ["#CIVIS name=",
     "#CIVIS name",
     "#CIVIS name # bad",
     "#CIVIS name= # really?"])
def test_parse_options_from_file_raises_error(line):
    fp = StringIO(line)
    with pytest.raises(ValueError):
        _parse_options_from_file(fp)


def test_parse_options_from_file():
    fp = StringIO("""\
#CIVIS name= my file name
#CIVIS pip = [numpy, scipy]
#CIVIS d={'cpu': 1024, 'memory': 8192}
#CIVIS namec= my file name # ugh
#CIVIS pipc = [numpy, scipy] # double ugh
#CIVIS dc={'cpu': 1024, 'memory': 8192} # comments might break me
""")
    params = _parse_options_from_file(fp)

    test_params = {
        'name': 'my file name',
        'pip': ['numpy', 'scipy'],
        'd': {'cpu': 1024, 'memory': 8192},
        'namec': 'my file name',
        'pipc': ['numpy', 'scipy'],
        'dc': {'cpu': 1024, 'memory': 8192}}
    assert params == test_params, "Parsed parameters are not correct!"


@pytest.mark.parametrize(
    'req',
    [{'cpu': '1024', 'memory': 2048.0, 'disk_space': 10},
     {'cpu': '1024', 'memory': 2048.0, 'diskSpace': 10}])
def test_cleanup_required_resources(req):
    expected_req = {'cpu': 1024, 'memory': 2048, 'disk_space': 10.0}
    test_req = _cleanup_required_resources(req)
    assert test_req == expected_req, (
        "Required resources were not cleaned up "
        "properly.")
    assert isinstance(test_req['cpu'], int), "CPU is not an int!"
    assert isinstance(test_req['memory'], int), "Memory is not an int!"
    assert isinstance(test_req['disk_space'], float), (
        "Disk space is not a float!")


@pytest.mark.parametrize(
    'old,new,res', [
        ({}, {'a': 5}, {'a': 5}),
        ({'a': 5}, {}, {'a': 5}),
        ({'a': 5}, {'a': 6}, {'a': 6}),
        ({'required_resources': {'a': 5}},
         {'required_resources': {'a': 6}},
         {'required_resources': {'a': 6}}),
        ({'required_resources': {'a': 5}},
         {},
         {'required_resources': {'a': 5}}),
        ({},
         {'required_resources': {'a': 6}},
         {'required_resources': {'a': 6}}),
        ({'arguments': {'a': 5}},
         {'arguments': {'a': 6}},
         {'arguments': {'a': 6}}),
        ({'arguments': {'a': 5}},
         {},
         {'arguments': {'a': 5}}),
        ({},
         {'arguments': {'a': 6}},
         {'arguments': {'a': 6}}),
        ({'params': [{'name': 'a', 'value': 10}]},
         {'params': [{'name': 'a', 'value': 11}]},
         {'params': [{'name': 'a', 'value': 11}]}),
        ({'params': [{'name': 'a', 'value': 10}, {'name': 'b', 'value': 11}]},
         {'params': [{'name': 'a', 'value': 11}]},
         {'params': [{'name': 'a', 'value': 11}, {'name': 'b', 'value': 11}]}),
        ({'params': [{'name': 'a', 'value': 10}]},
         {'params': [{'name': 'a', 'value': 11}, {'name': 'b', 'value': 11}]},
         {'params': [{'name': 'a', 'value': 11}, {'name': 'b', 'value': 11}]}),
        ({},
         {'params': [{'name': 'a', 'value': 11}]},
         {'params': [{'name': 'a', 'value': 11}]}),
        ({'params': [{'name': 'a', 'value': 10}]},
         {},
         {'params': [{'name': 'a', 'value': 10}]}),
        ({'files': ['a']}, {'files': ['b']}, {'files': ['a', 'b']}),
        ({}, {'files': ['b']}, {'files': ['b']}),
        ({'files': ['a']}, {}, {'files': ['a']})])
def test_join_container_params(old, new, res):
    _join_container_params(old, new)
    assert set(old) == set(res), "Parameters were not updated properly!"


@pytest.mark.parametrize(
    'params,res', [
        ({'files': ['a']}, {'files': ['a']}),
        ({}, {}),
        ({'files': 'a'}, {'files': ['a']}),
        ({'files': '   ,a,\n   '}, {'files': ['a']}),
        ({'files': '   ,a,\n   ,b,,  ,\n ,,'}, {'files': ['a', 'b']}),
        ({'files': '   a\n   '}, {'files': ['a']})])
def test_reformat_files_key(params, res):
    _reformat_files_key(params)
    assert params == res, "The 'files' key was not reformatted properly!"


@pytest.mark.parametrize('script_type', ['r', 'python'])
def test_process_options(client, script_type):
    params = {
        'a': 2, 'b': 3, 'required_resources': {'cpu': 8192.0},
        'files': 'a,   ,\n,,b'}
    cli_params = {
        'b': None, 'a': 4, 'c': 6, 'required_resources': {'memory': 10.0},
        'files': 'c,d'}

    test_params = _process_options(params, script_type, cli_params, client)

    expected_params = {
        'required_resources': {'cpu': 8192, 'memory': 10, 'disk_space': 16.0},
        'a': 4,
        'b': 3,
        'c': 6,
        'use_file_cache': False,
        'files': ['a', 'b', 'c', 'd']}

    if script_type == 'r':
        expected_params['docker_image_name'] = 'civisanalytics/datascience-r'
    else:
        expected_params['docker_image_name'] \
            = 'civisanalytics/datascience-python'
        expected_params['repo_cmd'] = 'python setup.py install'

    assert test_params == expected_params


@pytest.mark.parametrize('aws_cred_id', [None, 5])
def test_process_options_add_aws_creds(client, aws_cred_id):
    params = {
        'a': 2,
        'b': 3,
        'docker_image_name': 'p',
        'add_aws_creds': True,
        'aws_cred_id': aws_cred_id,
        'params': [{'name': 't', 'value': 5}]}
    cli_params = {'b': None, 'a': 4, 'c': 6}

    test_params = _process_options(params, 'r', cli_params, client)
    assert test_params == {
        'required_resources': {
            'cpu': 1024, 'memory': 8192, 'disk_space': 16.0},
        'a': 4,
        'b': 3,
        'c': 6,
        'docker_image_name': 'p',
        'use_file_cache': False,
        'arguments': {'AWS': aws_cred_id or 0},
        'params': [
            {'name': 't', 'value': 5},
            {'name': 'AWS', 'type': 'credential_aws'}]
        }, "Processed parameters are not right!"



from __future__ import print_function
from __future__ import division

import yaml

from .defaults import get_defaults


def _parse_options_from_file(fp):
    """Parse job options from comments in a file.

    The comments look like

        #CIVIS name=my job name

    which should get turned into

        {'name': 'my job name'}

    in the output dictionary.

    Parameters
    ----------
    fp : a file-like object
        A file-like object for reading the data from.

    Returns
    -------
    params : dict
        A dictionary of the parsed parameters. If nothing is found,
        an empty dictionary is returned.
    """
    params = {}
    for line in fp.readlines():
        if line[0:6] == '#CIVIS':
            # remove '#CIVIS', split on '=' and then strip out comments
            line = line[6:].strip()
            items = [i.strip() for i in line.split("=", 1)]
            if len(items) == 2:
                items[1] = items[1].split('#')[0].strip()

            if len(items) < 2 or len(items[1]) == 0:
                raise ValueError("civis-compute CLI option must be of form "
                                 "'#CIVIS key = value'!")

            # now parse the results with yaml
            params[items[0]] = yaml.load(items[1])

    return params


def _get_aws_cred_id(client):
    """Get the first AWS credential on platform."""
    name = client.username
    creds = client.credentials.list(type="Amazon Web Services S3")
    for cred in creds:
        if cred['owner'] == name:
            return cred['id']
    raise RuntimeError("Could not get AWS credential ID!")


def _add_aws_creds(params, client):
    """Add the AWS creds to the options.

    Builds a parameter called AWS
    that holds the AWS credentials stored on platform.
    """
    pname = 'AWS'

    if 'params' not in params:
        params['params'] = []

    params['params'].append(
        {'name': pname, 'type': 'credential_aws'})

    if 'arguments' not in params:
        params['arguments'] = {}

    params['arguments'][pname] \
        = params.pop('aws_cred_id', None) or _get_aws_cred_id(client)

    return params


def _cleanup_required_resources(req):
    """Convert tags to ints/floats for required_resources.

    cpu and memory should be ints
    disk_space should be a float
    diskSpace should be disk_space
    """
    for tag in ['cpu', 'memory']:
        req[tag] = int(req[tag])

    # this is always snake_case for inputs...
    if 'diskSpace' in req:
        req['disk_space'] = req['diskSpace']
        del req['diskSpace']

    req['disk_space'] = float(req['disk_space'])

    return req


def _join_container_params(old, new):
    """Join params w/ new updating old.

    This is like

        old.update(new)

    except that the keys

        required_resources
        params
        arguments
        files

    are updated instead of overwritten.

    Parameters
    ----------
    old : dict
        Dictionary of container parameters.
    new : dict
        Dictionary of container parameters.
    """

    old.update({
        k: v for k, v in new.items()
        if k not in ['required_resources', 'params', 'arguments', 'files']})

    # update the dict in old if it is there, otherwise just fill in
    for key in ['required_resources', 'arguments']:
        if key in new:
            if key in old:
                old[key].update(new[key])
            else:
                old[key] = new[key]

    # replace the items in the list by constructing dicts on the
    # name key
    # finally convert back to a list
    if 'params' in new:
        old_pars = {_p['name']: _p for _p in old.get('params', [])}
        old_pars.update({
            _p['name']: _p for _p in new.get('params', [])})
        old['params'] = [_p for _p in old_pars.values()]
        if len(old['params']) == 0:
            del old['params']

    if 'files' in new:
        if 'files' in old:
            old['files'].extend(new['files'])
        else:
            old['files'] = new['files']


def _reformat_files_key(params):
    """Reformat the 'files' key in params.

    Turns instances of the 'files' key as a comma-separated list of file
    names into a Python list of file names.

    The key is deleted if the resulting list of files is empty.

    Parameters
    ----------
    params : dict
    """
    if 'files' in params and isinstance(params['files'], str):
        new_files = params.pop('files', [])
        new_files = [
            i.strip()
            for i in new_files.split(',')
            if len(i.strip()) > 0]
        if new_files:
            params['files'] = new_files


def _process_options(file_params, script_type, cli_params, client):
    """Combine options from defaults, file and command line.

    Parameters
    ----------
    file_params : dict
        Options parsed from the file.
    script_type : str
        The type of script, either 'r' or 'python'.
    cli_params : dict
        Options from the command line.
    client : civis.APIClient
        An API client to use.

    Returns
    -------
    params : dict
        The final set of options.
    """
    # file > defaults always!
    params = get_defaults()
    _reformat_files_key(params)
    params['docker_image_name'] = params['docker_image_name'][script_type]
    if params['repo_cmd'].get(script_type):
        params['repo_cmd'] = params['repo_cmd'][script_type]
    else:
        del params['repo_cmd']
    _reformat_files_key(file_params)
    _join_container_params(params, file_params)

    # CLI > file always!
    # Remove None's to not set weird defaults.
    # click sets things to None if they are not passed, so ignore those
    _cli_params = {k: v for k, v in cli_params.items() if v is not None}
    _reformat_files_key(_cli_params)
    _join_container_params(params, _cli_params)

    # Add AWS credentials as environment variables in the container.
    if params.get('add_aws_creds', False):
        params = _add_aws_creds(params, client)
    params.pop('add_aws_creds', None)
    params.pop('aws_cred_id', None)

    # Cause yeah this happens...
    params['required_resources'] = _cleanup_required_resources(
        params['required_resources'])

    return params



from __future__ import print_function
from __future__ import division

from .conftest import mock

from ..utils import get_most_recent_run


@mock.patch('civiscompute.utils.civis.APIClient')
def test_get_most_recent_run(client_mock):
    client_mock.return_value.scripts.list_containers_runs.return_value = [
        {'finished_at': '2016-05-07', 'id': 0},
        {'finished_at': '2012-05-08', 'started_at': '2012-04-08', 'id': 1}]

    run_id = get_most_recent_run(-1)

    assert run_id == 0, "The most recent run was not returned!"
    client_mock.return_value.scripts.list_containers_runs.\
        assert_called_with(-1)


def test_get_most_recent_run_with_client():
    client_mock = mock.MagicMock()
    client_mock.scripts.list_containers_runs.return_value = [
        {'finished_at': '2010-05-07', 'id': 0},
        {'finished_at': '2012-01-08', 'started_at': '2012-04-08', 'id': 1}]

    run_id = get_most_recent_run(-1, client=client_mock)

    assert run_id == 1, "The most recent run was not returned!"
    client_mock.scripts.list_containers_runs.assert_called_with(-1)



from __future__ import print_function
from __future__ import division

import os
import sys
import hashlib
import datetime
import sqlite3
import dateutil.parser

import civis

from .defaults import FILEIDCACHE_FILE


class FileIDCache(object):
    """A persistent cache of file IDs on Civis Platform.

    Parameters
    ----------
    cache_file : str, optional
        The location to put the cache. The default location is
        in `civiscompute.defaults`.

    Methods
    -------
    list : list all the files in the cache
    get_fileid : get the file ID for a file
    """
    def __init__(self, client=None, cache_file=FILEIDCACHE_FILE):
        self._cache_file = os.path.expandvars(os.path.expanduser(cache_file))
        self._init_cache()
        self.client = civis.APIClient(resources='all')

    def _hash_func(self, fname):
        hash_md5 = hashlib.md5()
        with open(fname, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
        return hash_md5.hexdigest()

    def _init_cache(self):
        """setup the cache database"""
        if not os.path.exists(self._cache_file):
            pth, name = os.path.split(self._cache_file)
            # suppress any errors if the `~/.civiscompute` directory
            # already exists
            try:
                os.makedirs(pth)
            except OSError:
                pass
            init_db = True
        else:
            init_db = False

        self._conn = sqlite3.connect(self._cache_file, isolation_level=None)

        if init_db:
            self._conn.execute("CREATE TABLE FILES(\n"
                               "HASH TEXT PRIMARY KEY NOT NULL,\n"
                               "ID INT DEFAULT -1,\n"
                               "NAME TEXT DEFAULT NULL,\n"
                               "EXPDATE TEXT DEFAULT NULL);")

    def list(self):
        """List the files in the cache."""
        rows = self._conn.execute(
            'SELECT id, hash, expdate, name from files;').fetchall()
        fmt = "%- 25s %- 32s %- 32s %s"
        print(fmt % ("file id", "hash", "expiration date", "name"))
        for r in rows:
            print(fmt % r)

    def get_fileid(self, fname):
        """Get the file ID for a file.

        If the file is not in the cache or is going to expire in less than 14
        days, it is cached again.

        Parameters
        ----------
        fname : str
            The full path to the file.

        Returns
        -------
        file_id : int
            The file ID.
        """
        file_hash = self._hash_func(fname)
        d = self._conn.execute(
            "SELECT id, expdate from "
            "files where hash='%s';" % file_hash).fetchall()
        if len(d) == 0:
            file_id = self._insert_file(fname, file_hash)
        else:
            file_id = int(d[0][0])
            t_exp = d[0][1]

            # Re-upload if it is going to expire soon.
            if ((dateutil.parser.parse(t_exp) -
                 datetime.datetime.utcnow()) <
                    datetime.timedelta(days=14)):
                try:
                    with open(fname, 'rb') as fp:
                        file_id = civis.io.file_to_civis(
                            fp, 'file', client=self.client)

                    t_exp = self.client.files.get(file_id)['expires_at']

                    self._conn.execute(
                        "UPDATE files SET id=%d, expdate='%s' "
                        "WHERE hash='%s';" % (file_id, t_exp, file_hash))

                except Exception as e:
                    print(
                        "Could not upload file %s and put it into the "
                        "local cache: %s.%s: %s" % (
                            fname, e.__module__, e.__class__.__name__, e),
                        file=sys.stderr)
                    # Passing here since the file is in the files endpoint
                    # so we can still continue with the job
                    pass

        return file_id

    def _insert_file(self, fname, file_hash=None):
        """insert a file"""
        file_hash = file_hash or self._hash_func(fname)
        try:
            with open(fname, 'rb') as fp:
                file_id = civis.io.file_to_civis(
                    fp, 'file', client=self.client)

            t_exp = self.client.files.get(file_id)['expires_at']

            self._conn.execute(
                "INSERT INTO files VALUES ('%s', %d, '%s', '%s');" % (
                    file_hash, file_id, os.path.abspath(fname), t_exp))

        except Exception as e:
            print(
                "Could not upload file %s and put it into the "
                "local cache: %s.%s: %s" % (
                    fname, e.__module__, e.__class__.__name__, e),
                file=sys.stderr)
            sys.exit(-1)

        return file_id



from __future__ import print_function
from __future__ import division

import pytest

from ..parsing import _get_aws_cred_id, _add_aws_creds


def test_get_aws_cred_id(client):
    cid = _get_aws_cred_id(client)
    assert cid == 0, "The wrong credential was returned!"
    client.credentials.list.assert_called_with(
        type="Amazon Web Services S3")


def test_get_aws_cred_id_no_cred(client):
    client.credentials.list.return_value = [{'owner': 'b', 'id': 1}]

    with pytest.raises(RuntimeError):
        _get_aws_cred_id(client)

    client.credentials.list.assert_called_with(
        type="Amazon Web Services S3")


@pytest.mark.parametrize(
    'params', [
        {},
        {'params': [{'p1': 10}], 'arguments': {'p1': 5}},
        {'aws_cred_id': 10}])
def test_add_aws_creds(params, client):
    if params and 'params' in params:
        res = {'params': [{'p1': 10}], 'arguments': {'p1': 5}}
    else:
        res = {'params': [], 'arguments': {}}
    res['params'].append({'name': 'AWS', 'type': 'credential_aws'})
    res['arguments'].update({'AWS': params.get('aws_cred_id', 0)})

    test_params = _add_aws_creds(params, client)
    assert test_params == res, (
        "AWS credential parameter not added properly!")



from __future__ import print_function
from __future__ import division

import os
import sys
import civis

from .utils import get_most_recent_run


def _get(scriptid, runid=None, path=None):
    """Download job data for a script.

    During jobs, civis-compute automatically saves any outputs written
    to the directory given by the environment variable `${CIVIS_JOB_DATA}`.

    The `get` command downloads these outputs if they exist.

    The outputs are stored as a run output with the name

        civis_job_data_${CIVIS_JOB_ID}_${CIVIS_RUN_ID}

    Parameters
    ----------
    scriptid : int
        The container script ID.
    runid : int or None, optional
        The run ID to get outputs for. If None,
        the most recent run is used.
    path : str or None
        The path to download the data to. If None,
        the current working directory is used.
    """
    client = civis.APIClient(resources='all')

    try:
        runid = runid or get_most_recent_run(scriptid, client=client)
        outputs = client.scripts.list_containers_runs_outputs(scriptid, runid)

        for output in outputs:
            if (output['object_type'] == 'File' and
                    'civis_job_data_' in output['name']):

                if path:
                    oname = os.path.join(path, output['name'])
                else:
                    oname = output['name']

                print(oname)

                with open(oname, 'wb') as fp:
                    civis.io.civis_to_file(output['object_id'], fp)

                break
    except civis.base.CivisAPIError as e:
        print(
            "Could not get script %s outputs: %s.%s: %s" % (
                scriptid, e.__module__, e.__class__.__name__, e),
            file=sys.stderr)
        sys.exit(-1)



from __future__ import print_function
from __future__ import division

import pytest
import click
import civis
from .conftest import mock

from ..stat import (
    _print_script_runs,
    _print_script_logs,
    _print_scripts,
    _parse_user_ids,
    _status)


@pytest.mark.parametrize('hidden', [True, False])
@pytest.mark.parametrize('state', [None, 'running'])
@pytest.mark.parametrize('user_ids', [[1], []])
@mock.patch('civiscompute.stat.civis.APIClient')
def test_print_scripts(client_mock, capsys, state, user_ids, hidden):
    scripts = [
        {'id': 0,
         'name': 'n0',
         'author': {'id': 1, 'username': 'a'},
         'state': 'running',
         'last_run': None,
         'created_at': None},
        {'id': 1,
         'name': 'n1',
         'author': {'id': 2, 'username': 'b'},
         'state': 'running',
         'last_run': {'startedAt': None, 'finishedAt': None},
         'created_at': 'now'},
        {'id': 2,
         'name': 'n2',
         'author': {'id': 1, 'username': 'a'},
         'state': 'canceled',
         'last_run': {'startedAt': '-1', 'finishedAt': None},
         'created_at': 'like-right-now'},
        {'id': 3,
         'name': 'n3',
         'author': {'id': 2, 'username': 'b'},
         'state': 'canceled',
         'last_run': {'startedAt': '-2', 'finishedAt': '-3'},
         'created_at': 'yesterday'}]

    hidden_scripts = [
        {'id': 4,
         'name': 'n4',
         'author': {'id': 1, 'username': 'a'},
         'state': 'running',
         'last_run': {'startedAt': None, 'finishedAt': '-4'},
         'created_at': 'last-week'},
        {'id': 5,
         'name': 'n5',
         'author': {'id': 2, 'username': 'b'},
         'state': 'running',
         'last_run': {'startedAt': '-6', 'finishedAt': '-5'},
         'created_at': 'last-decade'},
        {'id': 6,
         'name': 'n6',
         'author': {'id': 1, 'username': 'a'},
         'state': 'canceled',
         'last_run': {'startedAt': '-8', 'finishedAt': '-7'},
         'created_at': 'sometime-in-the-last-day'},
        {'id': 7,
         'name': 'n7',
         'author': {'id': 2, 'username': 'b'},
         'state': 'canceled',
         'last_run': {'startedAt': '-10', 'finishedAt': '-9'},
         'created_at': 'yep-it-was'}]

    def _get_scripts(*args, **kwargs):
        if kwargs.get('hidden', False):
            _scripts = hidden_scripts
        else:
            _scripts = scripts

        if 'status' in kwargs and kwargs['status']:
            _scripts = [s for s in _scripts if s['state'] == kwargs['status']]
        if 'author' in kwargs and kwargs['author']:
            auths = [int(v) for v in kwargs['author'].strip().split(',')]
            _scripts = [s for s in _scripts if s['author']['id'] in auths]
        return _scripts

    client_mock.return_value.scripts.list = _get_scripts

    _print_scripts(user_ids, state, hidden)

    client_mock.assert_called_with(resources='all')

    out, err = capsys.readouterr()
    lines = out.split('\n')
    head = ("id                   author               status               "
            "created                   started                   finished"
            "                  name")
    assert lines[0].strip() == head, "Header for listing scripts is wrong!"

    all_scripts = scripts
    if hidden:
        all_scripts += hidden_scripts
    if state:
        all_scripts = [s for s in all_scripts if s['state'] == 'running']
    if user_ids:
        all_scripts = [s for s in all_scripts if s['author']['id'] == 1]

    for i, s in enumerate(all_scripts):
        items = lines[i+1].strip().split()
        assert items[0] == str(s['id']), (
            "Script id is wrong!")
        assert items[1] == s['author']['username'], (
            "Script username is wrong!")
        assert items[2] == s['state'], (
            "Script state is wrong!")
        assert items[3] == s['created_at'] or '-', (
            "Script created time is wrong!")
        if s['last_run']:
            assert items[4] == s['last_run']['startedAt'] or '-', (
                "Script started time is wrong!")
            assert items[5] == s['last_run']['finishedAt'] or '-', (
                "Script finished time is wrong!")
        else:
            assert items[4] == '-', (
                "Script started time is wrong!")
            assert items[5] == '-', (
                "Script finished time is wrong!")
        assert items[6] == s['name'], (
            "Script name is wrong!")


@pytest.mark.parametrize('runid', [2, None])
@mock.patch('civiscompute.stat.get_most_recent_run', autospec=True)
@mock.patch('civiscompute.stat.civis.APIClient')
def test_print_script_logs(client_mock, gmr_mock, capsys, runid):
    log_data = {
        'name': 'cool',
        'docker_command': 'echo 1 && \\\n  echo 2',
        'id': 4,
        'rand': 10}
    client_mock.return_value.scripts.get_containers.return_value = log_data
    gmr_mock.return_value = 5
    logs = [
        {'created_at': '3', 'message': 'c'},
        {'created_at': '2', 'message': 'b'},
        {'created_at': '1', 'message': 'a'}]
    client_mock.return_value.scripts.list_containers_runs_logs.return_value \
        = logs

    _print_script_logs(-1, runid=runid)

    client_mock.return_value.scripts.get_containers.assert_called_with(-1)
    if runid is None:
        assert gmr_mock.call_count == 1
    client_mock.return_value.scripts.list_containers_runs_logs.\
        assert_called_with(-1, runid or 5)

    out, err = capsys.readouterr()
    lines = out.split('\n')[:-1]  # remove extra blank line
    assert lines[0] == 'name: cool', "Script name is wrong!"
    assert lines[1] == 'id: 4', "Script ID is wrong!"
    assert lines[2] == 'docker_command:', (
        "Docker command head is wrong!")
    assert lines[3] == '  echo 1 && \\', (
        "First line of docker command is not right!")
    assert lines[4] == '    echo 2', (
        "Second line of docker command is not right!")
    assert lines[5] == 'rand: 10', (
        "Reamining key not printed!")
    assert lines[6] == 'log file:', (
        "Log file head was not printed!")
    assert lines[7] == '  [1] a', (
        "Log file line 1 was not printed!")
    assert lines[8] == '  [2] b', (
        "Log file line 2 was not printed!")
    assert lines[9] == '  [3] c', (
        "Log file line 3 was not printed!")
    assert len(lines) == 10, "Too many lines were printed!"


@mock.patch('civiscompute.stat.get_most_recent_run', autospec=True)
@mock.patch('civiscompute.stat.civis.APIClient')
def test_print_script_logs_no_runs(client_mock, gmr_mock, capsys):
    log_data = {
        'name': 'cool',
        'docker_command': 'echo 1 && \\\n  echo 2',
        'id': 4,
        'rand': 10}
    client_mock.return_value.scripts.get_containers.return_value = log_data
    gmr_mock.return_value = None
    logs = [
        {'created_at': '3', 'message': 'c'},
        {'created_at': '2', 'message': 'b'},
        {'created_at': '1', 'message': 'z'}]
    client_mock.return_value.scripts.list_containers_runs_logs.return_value \
        = logs

    _print_script_logs(-1)

    client_mock.return_value.scripts.get_containers.assert_called_with(-1)
    assert gmr_mock.call_count == 1
    assert client_mock.return_value.scripts.list_containers_runs_logs.\
        call_count == 0

    out, err = capsys.readouterr()
    lines = out.split('\n')[:-1]  # remove extra blank line
    assert lines[0] == 'name: cool', "Script name is wrong!"
    assert lines[1] == 'id: 4', "Script ID is wrong!"
    assert lines[2] == 'docker_command:', (
        "Docker command head is wrong!")
    assert lines[3] == '  echo 1 && \\', (
        "First line of docker command is not right!")
    assert lines[4] == '    echo 2', (
        "Second line of docker command is not right!")
    assert lines[5] == 'rand: 10', (
        "Reamining key not printed!")
    assert lines[6] == 'log file: -', (
        "Log file head was not printed!")
    assert len(lines) == 7, "Too many lines were printed!"


@mock.patch('civiscompute.stat.get_most_recent_run', autospec=True)
@mock.patch('civiscompute.stat.civis.APIClient')
def test_print_script_logs_raises(client_mock, gmr_mock):
    resp = mock.MagicMock()
    resp.content = None
    resp.reason = 'bad'
    resp.status_code = 502

    client_mock.return_value.scripts.get_containers.side_effect \
        = civis.base.CivisAPIError(resp)
    gmr_mock.return_value = 5

    with pytest.raises(SystemExit):
        _print_script_logs(-1)


@mock.patch('civiscompute.stat.civis.APIClient')
def test_print_script_runs(client_mock, capsys):
    run_data = [
        {'finished_at': '2016-05-07',
         'started_at': '2012-04-08',
         'id': 0,
         'state': 'happy',
         'error': None},
        {'finished_at': '2012-05-08',
         'started_at': '2012-04-02',
         'id': 1,
         'state': 'sad',
         'error': 'bad'}]
    client_mock.return_value.scripts.list_containers_runs.return_value \
        = run_data

    _print_script_runs(-1)

    client_mock.return_value.scripts.list_containers_runs.\
        assert_called_with(-1)

    out, err = capsys.readouterr()
    lines = out.split('\n')
    head = ("run id                    started                   finished"
            "                  status                    error msg")
    assert lines[0].strip() == head, "Run listing header is wrong!"

    for i in range(2):
        items = lines[i+1].strip().split()
        assert items[0] == str(run_data[i]['id']), (
            "run id is wrong for runs listing!")
        assert items[1] == run_data[i]['started_at'], (
            "starting time is wrong for runs listing!")
        assert items[2] == run_data[i]['finished_at'], (
            "finishing time is wrong for runs listing!")
        assert items[3] == run_data[i]['state'], (
            "state is wrong for runs listing!")
        err = run_data[i]['error'] or '-'
        assert ' '.join(items[4:]) == err, (
            "error is wrong for runs listing!")


@mock.patch('civiscompute.stat.civis.APIClient')
def test_print_script_runs_raises(client_mock, capsys):
    resp = mock.MagicMock()
    resp.content = None
    resp.reason = 'bad'
    resp.status_code = 502
    client_mock.return_value.scripts.list_containers_runs.side_effect \
        = civis.base.CivisAPIError(resp)

    with pytest.raises(SystemExit):
        _print_script_runs(-1)


@pytest.mark.parametrize('users', ['', 'me,you,everyone'])
def test_parse_user_ids(users):
    cv_mock = mock.MagicMock()
    cv_mock.users.list.side_effect = [[{'id': 1}], [{'id': 2}], [{'id': 3}]]

    ids = _parse_user_ids(users, cv_mock)

    if users:
        assert ids == [1, 2, 3], "The returned IDs are wrong!"
        cv_mock.users.list.assert_has_calls([
            mock.call(query='me'),
            mock.call(query='you'),
            mock.call(query='everyone')])
    else:
        assert ids == [], "The returned IDs are wrong!"
        assert cv_mock.call_count == 0


def test_status_raises():
    with pytest.raises(click.UsageError) as e:
        _status(-1, False, None, True, 10, False, False, 'all')

    assert '`--runs` or `--run-id`' in str(e.value), (
        "The wrong usage error was raised for runs and run IDs!")

    with pytest.raises(click.UsageError) as e:
        _status(-1, False, None, False, None, True, False, 'all')

    assert '`--last` or `scriptid`' in str(e.value), (
        "The wrong usage error was raised for script IDs and last!")


@pytest.mark.parametrize('mine', [True, False])
@pytest.mark.parametrize('user', [None, 'me,you'])
@pytest.mark.parametrize('hidden', [True, False])
@pytest.mark.parametrize('state', ['all', 'running'])
@mock.patch('civiscompute.stat._print_scripts', autospec=True)
@mock.patch('civiscompute.stat.civis')
def test_status_scripts(cv_mock, ps_mock, mine, user, hidden, state):
    cv_mock.APIClient.return_value.users.list_me.return_value = {'id': 1}
    cv_mock.APIClient.return_value.users.list.side_effect \
        = [[{'id': 2}], [{'id': 3}]]

    _status(None, mine, user, False, None, False, hidden, state)

    user_ids = []
    if mine:
        user_ids.append(1)

    if user:
        user_ids.extend([2, 3])

    if state != 'all':
        _state = state
    else:
        _state = None

    ps_mock.assert_called_with(
        user_ids,
        _state,
        hidden)

    if mine:
        assert cv_mock.APIClient.return_value.users.list_me.call_count == 1
    else:
        assert cv_mock.APIClient.return_value.users.list_me.call_count == 0


@mock.patch('civiscompute.stat._print_script_runs', autospec=True)
@mock.patch('civiscompute.stat.civis')
def test_status_script_runs(cv_mock, ps_mock):
    _status(-10, False, None, True, None, False, False, 'all')

    ps_mock.assert_called_with(-10)


@pytest.mark.parametrize('runid', [-11, None])
@mock.patch('civiscompute.stat._print_script_logs', autospec=True)
@mock.patch('civiscompute.stat.civis')
def test_status_script_logs(cv_mock, ps_mock, runid):
    _status(-10, False, None, False, runid, False, False, 'all')

    ps_mock.assert_called_with(-10, runid=runid)


@mock.patch('civiscompute.stat.LASTSCRIPTID_LOC', 'lastscript')
@mock.patch('civiscompute.stat._print_script_runs', autospec=True)
@mock.patch('civiscompute.stat.civis')
def test_status_last(cv_mock, ps_mock, cleandir):

    with open('lastscript', 'w') as fp:
        fp.write('-10')

    _status(None, False, None, True, None, True, False, 'all')

    ps_mock.assert_called_with(-10)


@mock.patch('civiscompute.stat.LASTSCRIPTID_LOC', 'lastscript')
@mock.patch('civiscompute.stat._print_script_runs', autospec=True)
@mock.patch('civiscompute.stat.civis')
def test_status_last_raises(cv_mock, ps_mock, cleandir):
    with pytest.raises(SystemExit):
        _status(None, False, None, True, None, True, False, 'all')



from __future__ import print_function
from __future__ import division

import pytest
import civis
from .conftest import mock

from ..cancel import _cancel


@mock.patch('civiscompute.cancel.civis.APIClient')
def test_cancel(client_mock):
    """Test canceling a job."""
    scriptids = [1]
    _cancel(scriptids)
    client_mock.return_value.scripts.post_cancel.assert_called_with(
        scriptids[0])


@mock.patch('civiscompute.cancel.civis.APIClient')
def test_post_cancel_raises(client_mock):
    """Test errors squashed on post cancel."""
    scriptids = [1]
    resp = mock.MagicMock()
    resp.content = None
    resp.reason = 'bad'
    resp.status_code = 502
    client_mock.return_value.scripts.post_cancel.side_effect \
        = civis.base.CivisAPIError(resp)

    with pytest.raises(SystemExit):
        _cancel(scriptids)

    client_mock.return_value.scripts.post_cancel.assert_called_with(
        scriptids[0])



from __future__ import print_function
from __future__ import division

import os
import six
import pkg_resources
import platform
import pytest
from civis.compat import TemporaryDirectory

if (six.PY2 or pkg_resources.parse_version(
        '.'.join(platform.python_version_tuple()[0:2]))
        == pkg_resources.parse_version('3.4')):
    import mock
else:
    from unittest import mock


@pytest.fixture
def client():
    client_mock = mock.MagicMock()
    client_mock.username = 'a'
    client_mock.credentials.list.return_value = [
        {'owner': 'a', 'id': 0},
        {'owner': 'b', 'id': 1}]

    return client_mock


@pytest.fixture()
def cleandir():
    cwd = os.getcwd()
    with TemporaryDirectory() as tmpdir:
        os.chdir(tmpdir)
        yield tmpdir
        os.chdir(cwd)


@pytest.fixture()
def example_notebook():
    return r"""{
"cells": [
{
"cell_type": "code",
"execution_count": null,
"metadata": {
"collapsed": true
},
"outputs": [],
"source": [
"import sys\n",
"\n",
"%matplotlib inline\n",
"    %matplotlib notebook"
]
}
],
"metadata": {
"kernelspec": {
"display_name": "Python 3",
"language": "python",
"name": "python3"
},
"language_info": {
"codemirror_mode": {
"name": "ipython",
"version": 3
},
"file_extension": ".py",
"mimetype": "text/x-python",
"name": "python",
"nbconvert_exporter": "python",
"pygments_lexer": "ipython3",
"version": "3.6.1"
}
},
"nbformat": 4,
"nbformat_minor": 2
}
"""



from __future__ import print_function
from __future__ import division

import nbformat
import nbconvert


def _reformat_notebook(script):
    """convert a notebook to a script and remove jupyter magics

    Parameters
    ----------
    script : str
        The path to the Jupyter notebook

    Returns
    -------
    code : str
        The notebook as a python script.
    """
    nb = nbformat.read(script, nbformat.NO_CONVERT)
    code = nbconvert.export(nbconvert.PythonExporter(), nb)[0]

    new_code = []
    for line in code.split('\n'):
        # Exclude magics - HACK - TODO

        # Matthew R. Becker 2/27/2017
        # So this one is tough. The following magics
        #
        #    %matplotlib inline
        #    %matplotlib notebook
        #
        # are only available in jupyter.
        # When notebooks are converted to scripts currently (2/17/2017 w/
        # version 5.1.1 of nbconvert), these magics are left in the notebook
        # as normal ipython magics. However, they cannot be executed by
        # ipython and cause an error. There is an open issue in nbconvert to
        # exclude these things
        # (see https://github.com/jupyter/nbconvert/pull/507). Until this
        # issue is addressed, we have to parse them out by hand.

        # The block of code here does the following.
        # 1) finds jupyter magics at any indentation level (they CAN be
        #    indented!)
        # 2) replaces them with `pass` in python (which does nothing) at the
        #    same indentation level
        if not (line.strip().startswith("get_ipython().magic("
                                        "'matplotlib") or
                line.strip().startswith("get_ipython().run_line_magic("
                                        "'matplotlib") or
                # python 2 - ugh
                line.strip().startswith("get_ipython().magic("
                                        "u'matplotlib") or
                line.strip().startswith("get_ipython().run_line_magic("
                                        "u'matplotlib")):
            new_code.append(line)
        else:
            new_code.append(line[:-len(line.lstrip())] + 'pass')
    return '\n'.join(new_code)



from __future__ import print_function
from __future__ import division

import sys
import civis


def _cancel(scriptids):
    """Cancel running scripts.

    Parameters
    ----------
    scriptids : tuple of ints
        The IDs of the scripts to cancel.
    """
    errored = False
    client = civis.APIClient(resources='all')
    for scriptid in scriptids:
        try:
            client.scripts.post_cancel(scriptid)
        except civis.base.CivisAPIError as e:
            errored = True
            print(
                "Could not cancel script %s: %s.%s: %s" % (
                    scriptid, e.__module__, e.__class__.__name__, e),
                file=sys.stderr)

    if errored:
        sys.exit(-1)



from __future__ import print_function
from __future__ import division

import os
import pytest
import civis
from .conftest import mock

from ..sub import (
    _upload_files,
    _build_container_cmd,
    _print_job,
    _post_and_run_job,
    _record_last_jobid,
    _sub_job,
    _sub)


@pytest.mark.parametrize('use_cache', [True, False])
@mock.patch('civiscompute.sub.civis.io.file_to_civis', autospec=True)
@mock.patch('civiscompute.sub.FileIDCache', autospec=True)
def test_upload_files(cache_mock, cv_mock, use_cache, cleandir):
    with open('a.txt', 'w') as fp:
        fp.write('a')

    with open('b.txt', 'w') as fp:
        fp.write('b')

    fnames = ['a.txt', 'b.txt']
    cache_mock.return_value.get_fileid.side_effect = [-1, -2]
    cv_mock.side_effect = [-3, -4]

    fileids = _upload_files(fnames, use_cache)

    if use_cache:
        assert fileids == {'a.txt': '-1', 'b.txt': '-2'}, (
            "fileids are wrong!")
        assert cache_mock.return_value.get_fileid.call_count == 2, (
            "FileIDCache not called!")
        assert cv_mock.call_count == 0, "file_to_civis called!"
    else:
        assert fileids == {'a.txt': '-3', 'b.txt': '-4'}, (
            "fileids are wrong!")
        assert cache_mock.return_value.get_fileid.call_count == 0, (
            "FileIDCache called!")
        assert cv_mock.call_count == 2, "file_to_civis not called!"


@mock.patch('civiscompute.sub.LASTSCRIPTID_LOC', 'a/lastid')
def test_record_last_jobid(cleandir):
    _record_last_jobid(-1)

    with open('a/lastid', 'r') as fp:
        data = fp.read()
    assert data.strip() == '-1', "Last job ID not written properly!"


@mock.patch('civiscompute.sub.LASTSCRIPTID_LOC', 'a/lastid')
@mock.patch('civiscompute.sub.open')
def test_record_last_jobid_squash(open_mock, cleandir):
    open_mock.side_effect = OSError("Bad thing!")
    _record_last_jobid(-1)
    assert not os.path.exists('a/lastid'), "Last job ID should not exist!"


def test_post_and_run_job():
    cv_mock = mock.MagicMock()
    cv_mock.scripts.post_containers.return_value = {'id': -1}
    req = -10
    container_cmd = -11
    docker_image = -12
    params = {'a': -13}

    retid = _post_and_run_job(
        req, container_cmd, docker_image, params, cv_mock)

    cv_mock.scripts.post_containers.assert_called_with(
        req, container_cmd, docker_image, **params)
    cv_mock.scripts.post_containers_runs.assert_called_with(-1)
    assert retid == -1, "Wrong job ID returned!"


@pytest.mark.parametrize('fail_post', [True, False])
def test_post_and_run_job_raises(fail_post, capsys):
    cv_mock = mock.MagicMock()
    resp = mock.MagicMock()
    resp.content = None
    resp.reason = 'bad stuff'
    resp.status_code = 5018
    if fail_post:
        cv_mock.scripts.post_containers.side_effect \
            = civis.base.CivisAPIError(resp)
    else:
        cv_mock.scripts.post_containers.return_value = {'id': -1}
        cv_mock.scripts.post_containers_runs.side_effect \
            = civis.base.CivisAPIError(resp)
    req = -10
    container_cmd = -11
    docker_image = -12
    params = {'a': -13}

    with pytest.raises(SystemExit):
        _post_and_run_job(req, container_cmd, docker_image, params, cv_mock)

    cv_mock.scripts.post_containers.assert_called_with(
        req, container_cmd, docker_image, **params)

    _, err = capsys.readouterr()
    if fail_post:
        assert cv_mock.scripts.post_containers_runs.call_count == 0
        assert "Could not create container" in err, (
            "Stderr is not right for failed create!")
    else:
        cv_mock.scripts.post_containers_runs.assert_called_with(-1)
        assert "Could not run container -1" in err, (
            "Stderr is not right for failed run!")


def test_print_job(capsys):
    req = {'cpu': 10, 'memory': 11, 'disk_space': 12}
    container_cmd = 'blah && \\\nblah && \\\nblah'
    docker_image = 'containers!'
    params = {'a': 13}

    _print_job(req, container_cmd, docker_image, params)

    out, err = capsys.readouterr()
    assert err == '', "Stderr was produced!"

    test_out = """\
container script config:
  a: 13
  docker_image_name: containers!
  required_resources:
    cpu: 10
    disk_space: 12
    memory: 11
container script cmd:
  blah && \\
  blah && \\
  blah
"""
    assert out == test_out, "Stdout is wrong!"


@pytest.mark.parametrize('stype', ['py', 'ipynb', 'R', 'r', 'b', 'sh', None])
@pytest.mark.parametrize('cli_args', [('a', 'b'), tuple()])
@pytest.mark.parametrize('fileids', [{}, {'a/f1': -1, 'b/f2': -2, 'a/f3': -3}])
@pytest.mark.parametrize('repo_cmd', [None, 'pip me'])
@pytest.mark.parametrize('shell_cmd', [None, 'sh me'])
def test_build_container_cmd(
        stype, cli_args, fileids, repo_cmd, shell_cmd, cleandir):

    if stype:
        script = 's' + '.' + stype
        with open(script, 'w') as fp:
            fp.write('blah')

        if stype == 'py':
            remote_script = 'rs.py'
        else:
            remote_script = script
    else:
        script = 's'
        remote_script = script

    cmd = _build_container_cmd(
        script, remote_script, cli_args, fileids, repo_cmd, shell_cmd)

    lines = cmd.split('\n')

    if cli_args:
        arg_str = ' a b '
    else:
        arg_str = ' '
    if stype == 'py':
        cmd = 'python '
    elif stype == 'r' or stype == 'R':
        cmd = 'Rscript '
    elif stype == 'ipynb':
        cmd = "ipython --InteractiveShell.colors='nocolor' "
    elif stype:
        cmd = 'sh '
    else:
        cmd = ''
    run_str = '%s%s%s&& \\' % (cmd, remote_script, arg_str)
    assert run_str in lines

    if stype:
        assert 'chmod a+rwx %s && \\' % remote_script in lines

    if repo_cmd:
        assert 'cd /app && %s && cd .. && \\' % repo_cmd in lines
        # repo command comes before run
        assert (
            lines.index('cd /app && %s && cd .. && \\' % repo_cmd) <
            lines.index(run_str))

    if shell_cmd:
        assert '%s && \\' % shell_cmd in lines

    # shell command comes before the repo command
    if shell_cmd and repo_cmd:
        assert (
            lines.index('%s && \\' % shell_cmd) <
            lines.index('cd /app && %s && cd .. && \\' % repo_cmd))

    if fileids:
        assert 'mkdir -p a && \\' in lines
        assert 'mkdir -p b && \\' in lines
        assert 'civis files download -1 a/f1 && \\' in lines
        assert 'civis files download -2 b/f2 && \\' in lines
        assert 'civis files download -3 a/f3 && \\' in lines

        # have to make dirs before we download
        assert (
            lines.index('mkdir -p a && \\') <
            lines.index('civis files download -1 a/f1 && \\'))
        assert (
            lines.index('mkdir -p b && \\') <
            lines.index('civis files download -2 b/f2 && \\'))
        assert (
            lines.index('mkdir -p a && \\') <
            lines.index('civis files download -3 a/f3 && \\'))

        # all files should come before any shell or repo
        if shell_cmd:
            assert (
                lines.index('civis files download -3 a/f3 && \\') <
                lines.index('%s && \\' % shell_cmd))

        if repo_cmd:
            assert (
                lines.index('civis files download -3 a/f3 && \\') <
                lines.index('cd /app && %s && cd .. && \\' % repo_cmd))


@pytest.mark.parametrize('script', ['script.ipynb', 'script.py'])
@mock.patch('civiscompute.sub.TemporaryDirectory', autospec=True)
@mock.patch('civiscompute.sub.civis.APIClient', autospec=True)
@mock.patch('civiscompute.sub._record_last_jobid', autospec=True)
@mock.patch('civiscompute.sub._sub_job', autospec=True)
def test_sub(mock_sub, mock_rcd, mock_cv, mock_tmp,
             script, cleandir, example_notebook):
    cli_args = ['1', '2']
    dry_run = False
    cli_params = {'a': 5, 'dry_run': dry_run}
    mock_sub.return_value = 5
    mock_tmp.return_value.__enter__.return_value = 'tmp'
    os.makedirs('tmp')

    with open(script, 'w') as fp:
        if script.endswith('.ipynb'):
            fp.write(example_notebook)
        else:
            fp.write('test')

    _sub(script, cli_args, cli_params)

    if script.endswith('.ipynb'):
        mock_sub.assert_called_with(
            os.path.join('tmp', script),
            script,
            cli_args,
            cli_params,
            mock_cv.return_value,
            dry_run)
    else:
        mock_sub.assert_called_with(
            script,
            script,
            cli_args,
            cli_params,
            mock_cv.return_value,
            dry_run)
    mock_rcd.assert_called_with(5)


@pytest.mark.parametrize('repo', [None, 'gitrepo'])
@pytest.mark.parametrize('dry_run', [True, False])
@pytest.mark.parametrize('script', ['script.r', 'script.py', 'echo 1'])
@mock.patch('civiscompute.sub.civis.io.file_to_civis', autospec=True)
@mock.patch('civiscompute.sub._post_and_run_job', autospec=True)
def test_sub_job(mock_sub, mock_ftc, script, dry_run, repo, cleandir):

    if script.endswith('.r') or script.endswith('.py'):
        with open(script, 'w') as fp:
            fp.write('#CIVIS files=test.txt\n')
            fp.write('#CIVIS name=my script\n')

        with open('test.txt', 'w') as fp:
            fp.write('hi!\n')

    with open('test2.txt', 'w') as fp:
        fp.write('hi again!')

    cli_args = ('a', 'b', 'test2.txt')
    cli_params = {
        'required_resources': {'cpu': 567},
        'shell_cmd': 'what'}
    if repo:
        cli_params['repo_http_uri'] = repo
    client = mock.MagicMock()
    mock_sub.return_value = 5
    mock_ftc.side_effect = [6, 7, 8]

    if script.endswith('.r'):
        remote_script = 'blah.r'
    else:
        remote_script = script

    jobid = _sub_job(
        script, remote_script, cli_args, cli_params, client, dry_run)

    if dry_run:
        # make sure no jobs are submitted
        assert jobid is None
        assert mock_sub.call_count == 0
    else:
        # make sure job is submitted only once
        assert jobid == 5
        assert mock_sub.call_count == 1

        # check arguments for job submission
        args, _ = mock_sub.call_args

        # resources
        assert args[0] == {'cpu': 567, 'memory': 8192, 'disk_space': 16.0}

        # docker image name
        if script.endswith('.r'):
            assert args[2] == 'civisanalytics/datascience-r'
        elif script.endswith('.py'):
            assert args[2] == 'civisanalytics/datascience-python'

        # parameters
        test_params = {}
        if script.endswith('.r') or script.endswith('.py'):
            test_params['name'] = 'my script'
        if repo:
            test_params['repo_http_uri'] = repo
        assert args[3] == test_params

        # make sure repo_cmd and repo_http_uri were processed correctly
        if repo and script.endswith('.py'):
            assert 'python setup.py install' in args[1]

        # did the shell_cmd get through?
        assert 'what &&' in args[1]

        # make sure files are downloaded into the container
        if script.endswith('.r') or script.endswith('.py'):
            assert 'civis files download 6' in args[1]
            assert 'civis files download 7' in args[1]
            assert 'civis files download 8' in args[1]
        else:
            assert 'civis files download 6' in args[1]

        # check command invocation
        if script.endswith('.r'):
            assert 'Rscript blah.r a b test2.txt' in args[1]
        elif script.endswith('.py'):
            assert 'python script.py a b test2.txt' in args[1]
        else:
            assert script + ' a b test2.txt' in args[1]

    # make sure files are uploaded properly
    if dry_run:
        assert mock_ftc.call_count == 0
    elif script.endswith('.r') or script.endswith('.py'):
        assert mock_ftc.call_count == 3
    else:
        assert mock_ftc.call_count == 1



from __future__ import print_function
from __future__ import division

import os
import pytest
import civis
from .conftest import mock

from ..get import _get


@pytest.mark.parametrize('runid', [None, 3])
@pytest.mark.parametrize('path', [None, 'pth'])
@pytest.mark.parametrize('no_output', [False, True])
@pytest.mark.parametrize('oname', ['civis_job_data_f', 'fff'])
@mock.patch('civiscompute.get.civis.io')
@mock.patch('civiscompute.get.open')
@mock.patch('civiscompute.get.civis.io.civis_to_file', autospec=True)
@mock.patch('civiscompute.get.get_most_recent_run', autospec=True)
@mock.patch('civiscompute.get.civis.APIClient')
def test_get(
        client_mock, gmr_mock, ctf_mock, open_mock, civisio_mock,
        runid, path, no_output, oname):
    """Test getting job outputs."""

    scriptid = 1
    outputs = [{'object_type': 'RANDOM', 'name': 'filename8', 'object_id': 8}]
    if not no_output:
        outputs.append(
            {'object_type': 'File', 'name': oname, 'object_id': 7})

    gmr_mock.return_value = runid or 2
    client_mock.return_value.scripts.list_containers_runs_outputs.\
        return_value = outputs
    open_mock.return_value.__enter__.return_value = 10

    _get(scriptid, runid=runid, path=path)

    # assert that the correct runid is used, if specified
    if runid is None:
        gmr_mock.assert_called_with(scriptid, client=client_mock.return_value)
    else:
        assert gmr_mock.call_count == 0

    # assert that the proper job run outputs are accessed
    client_mock.return_value.scripts.list_containers_runs_outputs.\
        assert_called_with(scriptid, runid or 2)

    # assert that if outputs are present and have the right name,
    # they are downloaded
    if not no_output and 'civis_job_data_' in oname:
        if path:
            oname = os.path.join(path, oname)
        open_mock.assert_called_with(oname, 'wb')
        civisio_mock.civis_to_file.assert_called_with(7, 10)
    else:
        assert open_mock.call_count == 0
        assert civisio_mock.civis_to_file.call_count == 0


@mock.patch('civiscompute.get.civis.APIClient')
def test_get_raises(client_mock):
    """Test errors squashed for get."""
    scriptids = [1]
    resp = mock.MagicMock()
    resp.content = None
    resp.reason = 'bad'
    resp.status_code = 502
    client_mock.return_value.scripts.list_containers_runs.side_effect \
        = civis.base.CivisAPIError(resp)

    with pytest.raises(SystemExit):
        _get(scriptids)



from __future__ import print_function
from __future__ import division

import dateutil.parser
import civis


def get_most_recent_run(scriptid, client=None):
    """Get the most recent run of a container script.

    Parameters
    ----------
    scriptid : int
        The ID of the container script.
    client : None or civis.APIClient, optional
        A Civis API client to use. If not given, one will be
        instantiated.

    Returns
    -------
    runid : int
        The ID of the latest run.
    """
    client = client or civis.APIClient(resources='all')

    # First get all of the runs.
    runs = client.scripts.list_containers_runs(scriptid)

    if len(runs) > 0:
        # Get the most recent one.
        run_times = [dateutil.parser.parse(r['finished_at'] or r['started_at'])
                     for r in runs]
        run = runs[run_times.index(max(run_times))]
        return int(run['id'])
    else:
        return None



from __future__ import print_function
from __future__ import division

import os
import sys

import click
import yaml

from .cancel import _cancel
from .caching import FileIDCache
from .get import _get
from .stat import _status
from .sub import _sub
from .defaults import FILEIDCACHE_FILE


# parameter type for YAML text
class YAMLParamType(click.ParamType):
    """
    A click parameter type for YAML/JSON strings.
    See http://click.pocoo.org/5/parameters/#implementing-custom-types.
    """

    name = 'YAMLTEXT'

    def convert(self, value, param, ctx):
        if value is None:
            return value

        try:
            result = yaml.load(value)
            return result
        except Exception:
            self.fail(
                "Could not load YAML from string: %s" % value, param, ctx)


YAML = YAMLParamType()


@click.group()
def cli():
    """Welcome to the civis-compute command line interface!

    Make sure to have your Civis API key in the local environment as
    `CIVIS_API_KEY`.
    """
    pass


@cli.command()
@click.argument('script')
@click.argument('args', nargs=-1)
@click.option('--name', type=str, help="Name of the container.")
@click.option('--parent-id', type=int,
              help="ID of the parent job that will trigger this script.")
@click.option('--user-context', type=str,
              help='Who to execute the script as when run as a template. '
                   'Either "runner" or "author".')
@click.option('--params', type=YAML,
              help="Definition of the parameters this script accepts in the "
                   "arguments field. See the Civis API docs.")
@click.option('--arguments', type=YAML,
              help="Dictionary of name/value pairs to use to run this script. "
                   "Only settable if this script has defined params. See the "
                   "Civis API docs.")
@click.option('--schedule', type=YAML,
              help="Schedule of when this script will be run. "
                   "See the Civis API docs.")
@click.option('--notifications', type=YAML,
              help="Notifications to send after the script is run. "
                   "See the Civis API docs.")
@click.option('--required-resources', type=YAML,
              help="Resources to allocate for the container. "
                   "This parameter should be set to a string "
                   "with YAML specifying the required "
                   "resources (e.g., \"{'cpu': 2048}\"). "
                   "Possible keys are 'cpu', 'memory' and "
                   "'disk_space'. See the Civis API docs.")
@click.option('--repo-http-uri', type=str,
              help="Location of a github repo to clone into the "
                   "container, e.g. github.com/my-user/my-repo.git.")
@click.option('--repo-ref', type=str,
              help="Tag or branch of the github repo to clone "
                   "into the container.")
@click.option('--remote-host-credential-id', type=int,
              help="ID of the database credentials to pass into "
                   "the environment of the container.")
@click.option('--git-credential-id', type=int,
              help="ID of the git credential to be used when "
                   "checking out the specified git repo. If not "
                   "supplied, the first git credential you've "
                   "submitted will be used. Unnecessary if no "
                   "git repo is specified or the git repo "
                   "is public.")
@click.option('--docker-image-name', type=str,
              help="Name of the docker image to pull from DockerHub.")
@click.option('--docker-image-tag', type=str,
              help="Tag of the docker image to pull from DockerHub. "
                   "(default: latest)")
@click.option('--time-zone', type=str, help="Time zone of this script.")
@click.option('--hidden/--no-hidden', default=None,
              help="Hidden status of the object. Setting this to true hides "
                   "it from most API endpoints. The object can still be "
                   "queried directly by ID")
@click.option('--target-project-id', type=int,
              help="ID of the target project to which script outputs "
                   "will be added.")
@click.option('--cancel-timeout', type=int,
              help="If non-zero, amount of time in seconds to wait between "
                   "sending SIGTERM and SIGKILL to cancel a script. "
                   "Must be in the range 0 to 60.")
@click.option('--files', type=str,
              help="File or comma-separated list of files to upload "
                   "into the container. These files will be placed at "
                   "the same path in the container as they are locally.")
@click.option('--repo-cmd', type=str,
              help="Command to use to process the cloned repo, "
                   "if applicable. (default: `python setup.py install`)")
@click.option('--use-file-cache/--no-use-file-cache', default=None,
              help="Use the local cache of files endpoint IDs to avoid "
                   "duplicate file uploads.")
@click.option('--add-aws-creds/--no-add-aws-creds', default=None,
              help="Add AWS credentials to the submitted script. "
                   "The first AWS credential found is used.")
@click.option('--aws-cred-id', type=int, default=None,
              help="ID of the AWS credential to use when `add-aws-creds` "
                   "is set. If this is not provided, then the first returned "
                   "by the API will be used.")
@click.option('--shell-cmd', type=str,
              help="Shell commands to be executed after all data is "
                   "downloaded to the container but before any other "
                   "installation steps. These can be used to, e.g., install "
                   "packages in a custom way, move data around in the "
                   "container, etc.")
@click.option('--dry-run', is_flag=True,
              help="Configure the container, but do not submit it and "
                   "instead print the configuration.")
def submit(script, args, **kwargs):
    """Submit a SCRIPT to Civis Platform.

    Arguments can be passed directly as ARGS.

    See the Civis API docs

        https://platform.civisanalytics.com/api#v1_post_scripts_containers

    for information about the options.

    Make sure to have your Civis API key in the local environment as
    `CIVIS_API_KEY`.
    """
    _sub(script, args, kwargs)


@cli.command()
@click.argument('scriptids', nargs=-1)
def cancel(scriptids):
    """Cancel Civis Platform container scripts.

    Make sure to have your Civis API key in the local environment as
    `CIVIS_API_KEY`.
    """
    _cancel(scriptids)


@cli.group()
def cache():
    """Manage the cache of file IDs.

    Make sure to have your Civis API key in the local environment as
    `CIVIS_API_KEY`.
    """
    pass


@cache.command()
def list():
    """List items in the file ID cache."""
    fc = FileIDCache()
    fc.list()


@cache.command()
def clear():
    """Clear the file ID cache."""
    fname = os.path.expandvars(os.path.expanduser(FILEIDCACHE_FILE))
    try:
        os.remove(fname)
    except Exception as e:
        print(
            "Could not clear the cache at %s: %s.%s: %s" % (
                    fname, e.__module__, e.__class__.__name__, e),
            file=sys.stderr)
        sys.exit(-1)


@cli.command()
@click.argument('scriptid', nargs=1)
@click.argument('path', type=str, default='.', nargs=1)
@click.option(
    '--run-id',
    'runid',
    type=int,
    help="Get output for a specific run RUN_ID instead of the latest run.")
def get(scriptid, path, runid):
    """Download the outputs for a given SCRIPTID.

    Before the container script finishes running, civis-compute automatically
    saves any outputs written to the directory given by the environment
    variable `${CIVIS_JOB_DATA}` as a Civis Platform file.

    The `get` command downloads these outputs if they exist.

    The most recent run is used unless a run is specified via `--run-id`.

    Optionally download to the given PATH. Otherwise use the current
    working directory.

    Make sure to have your Civis API key in the local environment as
    `CIVIS_API_KEY`.
    """
    _get(scriptid, runid=runid, path=path)


@cli.command()
@click.argument('scriptid', nargs=-1)
@click.option('--mine', '-m', default=False, is_flag=True,
              help="List only your scripts.")
@click.option('--user', type=str, default=None,
              help="List scripts only from this user (or users in a "
                   "comma-separated list).")
@click.option('--runs', default=False, is_flag=True,
              help="List all runs for a given SCRIPTID if SCRIPTID is given.")
@click.option('--run-id', 'runid', type=int, default=None,
              help="List info for a given RUNID if SCRIPTID is given.")
@click.option('--last', '-l', 'last', is_flag=True,
              help="Print details of your last submitted script as if "
                   "you had given it as SCRIPTID.")
@click.option('--hidden', is_flag=True,
              help='List hidden scripts too.')
@click.option('--running', '-r', 'state', flag_value='running', default=True,
              help='List only running scripts. (default)')
@click.option('--all', '-a', 'state', flag_value='all',
              help='List scripts in all states.')
@click.option('--canceled', '-c', 'state', flag_value='canceled',
              help='List only canceled scripts.')
@click.option('--succeeded', '-s', 'state', flag_value='succeeded',
              help='List only succeeded scripts.')
@click.option('--failed', '-f', 'state', flag_value='failed',
              help='List only failed scripts.')
@click.option('--idle', '-i', 'state', flag_value='idle',
              help='List only idle scripts.')
def status(scriptid, mine, user, runs, runid, last, hidden, state):
    """Inspect Civis Platform container scripts.

    Specify SCRIPTID to inspect a given script. Otherwise the status of
    the civis-compute queue will be returned. Only up to ~50 scripts can be
    displayed and only scripts of type "Container" are returned (this includes
    custom and templated scripts).

    Make sure to have your Civis API key in the local environment as
    `CIVIS_API_KEY`.
    """

    _status(
        int(scriptid[0]) if len(scriptid) > 0 else None,
        mine, user, runs, runid, last, hidden, state)



import os
from setuptools import find_packages, setup


def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as _in:
        return _in.read()


__version__ = None
exec(read('civiscompute/version.py'))

setup(
    name="civis-compute",
    version=__version__,
    author="Civis Analytics Inc",
    author_email="opensource@civisanalytics.com",
    url="https://www.civisanalytics.com",
    description="Batch computing in the cloud with Civis Platform.",
    packages=find_packages(),
    long_description=read('README.rst'),
    include_package_data=True,
    license="BSD-3",
    install_requires=read('requirements.txt').strip().split('\n'),
    entry_points={
        'console_scripts': ['civis-compute = civiscompute.cli:cli']})



from __future__ import print_function
from __future__ import division

import os
import yaml

FILEIDCACHE_FILE = '~/.civiscompute/fileidcache.db'
LASTSCRIPTID_LOC = '~/.civiscompute/lastscriptid'


def get_defaults():
    defaults = {
        'use_file_cache': False,
        'required_resources': {
            'cpu': 1024, 'memory': 8192, 'disk_space': 16.0},
        'docker_image_name': {
            'python': 'civisanalytics/datascience-python',
            'r': 'civisanalytics/datascience-r'},
        'repo_cmd': {
            'python': 'python setup.py install',
            'r': None},
        'add_aws_creds': False,
        'aws_cred_id': None}

    config_loc = os.path.expanduser('~/.civiscompute/config.yml')
    if os.path.exists(config_loc):
        with open(config_loc, 'r') as fp:
            defaults.update(yaml.load(fp))

    return defaults



from __future__ import print_function
from __future__ import division

import os
from civis.compat import TemporaryDirectory

from ..notebooks import _reformat_notebook


def test_reformat_notebook(example_notebook):
    with TemporaryDirectory() as tmpdir:
        with open(os.path.join(tmpdir, 'script.ipynb'), 'w') as fp:
            fp.write(example_notebook)
        code = _reformat_notebook(os.path.join(tmpdir, 'script.ipynb'))

    assert code == """
# coding: utf-8

# In[ ]:


import sys

pass
    pass

""", "Notebook was not reformatted correctly!"



from __future__ import print_function
from __future__ import division

import os
import hashlib

import datetime
import pytest
from civis.compat import TemporaryDirectory
from .conftest import mock

from ..caching import FileIDCache


@pytest.mark.parametrize("day_offset", [20, 1])
@mock.patch('civiscompute.caching.civis.io')
@mock.patch('civiscompute.caching.civis.APIClient')
def test_cache(client_mock, civisio_mock, day_offset):
    """Make sure the file ID cache works."""

    with TemporaryDirectory() as dirname:
        test_name1 = os.path.join(dirname, 'test1.txt')
        with open(test_name1, 'w') as fp:
            fp.write('a')

        test_name2 = os.path.join(dirname, 'test2.txt')
        with open(test_name2, 'w') as fp:
            fp.write('b')

        exp_time = str(datetime.datetime.utcnow() +
                       datetime.timedelta(days=day_offset))
        exp_time30 = str(datetime.datetime.utcnow() +
                         datetime.timedelta(days=30))

        civisio_mock.file_to_civis.side_effect = [-1, -2, -3]
        client_mock().files.get.return_value = {'expires_at': exp_time}

        # Upload once.
        fc = FileIDCache(cache_file=os.path.join(dirname, 'cache.db'))
        fid = fc.get_fileid(test_name1)
        assert fid == -1, "File ID for cache insert is wrong!"
        client_mock().files.get.assert_called_with(-1)
        assert civisio_mock.file_to_civis.call_count == 1

        fid = fc.get_fileid(test_name2)
        assert fid == -2, "File ID for cache insert is wrong!"
        client_mock().files.get.assert_called_with(-2)
        assert civisio_mock.file_to_civis.call_count == 2

        # Get it back out.
        client_mock.reset_mock()
        if day_offset < 14:
            client_mock().files.get.return_value = {'expires_at': exp_time30}
        fid = fc.get_fileid(test_name1)

        # If it is going to expire, make sure we uploaded it again
        if day_offset >= 14:
            assert fid == -1, "File ID for cache get is wrong!"
            assert civisio_mock.file_to_civis.call_count == 2
            assert client_mock().files.get.call_count == 0
        else:
            assert fid == -3, "File ID for cache get is wrong!"
            assert civisio_mock.file_to_civis.call_count == 3
            client_mock().files.get.assert_called_with(-3)


@mock.patch('civiscompute.caching.civis.io')
@mock.patch('civiscompute.caching.civis.APIClient')
def test_cache_list(client_mock, civisio_mock, capsys):
    """Make sure the file ID cache listing works."""

    with TemporaryDirectory() as dirname:
        test_name1 = os.path.join(dirname, 'test1.txt')
        with open(test_name1, 'w') as fp:
            fp.write('a')
        hash_test_name = hashlib.md5('a'.encode('utf-8')).hexdigest()

        test_name2 = os.path.join(dirname, 'test2.txt')
        with open(test_name2, 'w') as fp:
            fp.write('b')
        hash_test_name2 = hashlib.md5('b'.encode('utf-8')).hexdigest()

        exp_time = str(datetime.datetime.utcnow() +
                       datetime.timedelta(days=15))

        civisio_mock.file_to_civis.side_effect = [-1, -2]
        client_mock().files.get.return_value = {'expires_at': exp_time}

        # Upload some files.
        fc = FileIDCache(cache_file=os.path.join(dirname, 'cache.db'))
        fc.get_fileid(test_name1)
        fc.get_fileid(test_name2)

        # List the cahce and make sure that looks OK.
        fc.list()
        out, err = capsys.readouterr()
        print(out)
        lines = out.split('\n')
        assert lines[0].strip() == \
            ("file id                   hash                             "
             "expiration date                  name"), (
             "File cache list header is wrong!")

        items = lines[1].strip().split()
        assert items[0] == str(-1), "file id is wrong for cache listing!"
        assert items[1] == hash_test_name, "hash is wrong for cache listing!"
        assert ' '.join(items[2:4]) == exp_time, (
            "expiration time is wrong for cache listing!")
        assert items[4] == test_name1, "name is wrong for cache listing!"

        items = lines[2].strip().split()
        assert items[0] == str(-2), "file id is wrong for cache listing!"
        assert items[1] == hash_test_name2, "hash is wrong for cache listing!"
        assert ' '.join(items[2:4]) == exp_time, (
            "expiration time is wrong for cache listing!")
        assert items[4] == test_name2, "name is wrong for cache listing!"



from __future__ import print_function
from __future__ import division

import os
import sys
from civis.compat import TemporaryDirectory

import yaml
import civis

from .defaults import LASTSCRIPTID_LOC
from .parsing import _parse_options_from_file, _process_options
from .notebooks import _reformat_notebook
from .caching import FileIDCache


def _upload_files(files, use_cache):
    """Upload file(s) to files endpoint.

    Parameters
    ----------
    files : list of str
        List of files to upload.
    use_cache : bool
        If True, cache the file IDs locally. Otherwise files will
        always be uploaded.

    Returns
    -------
    fileids : dict
        Dictionary mapping file names to Civis Platform file IDs as strings.
    """
    if use_cache:
        cache = FileIDCache()

    fileids = {}
    for fname in files:
        if use_cache:
            fileid = cache.get_fileid(fname)
        else:
            with open(fname, 'rb') as fp:
                fileid = civis.io.file_to_civis(fp, 'file')
        fileids[fname] = str(fileid)
    return fileids


def _build_container_cmd(
        script, remote_script, cli_args, fileids, repo_cmd, shell_cmd):
    """Build the docker command.

    Parameters
    ----------
    script : str
        The /path/to/the/script.XXX or a shell command.
    remote_script : str
        The path to put the script in the container. This could differ
        from the local path for notebooks since they are reformatted
        locally before being executed.
    cli_args : list
        List of arguments from the command line.
    fileids : dict
        Dictionary mapping file names to file IDs.
    repo_cmd : str or None
        Command to use to install the cloned repo in the container. If None
        no repo is installed.
    shell_cmd : str or None
        A shell command to execute before the repo is installed (possibly)
        and after all files from the file IDs are downloaded to the
        container.

    Returns
    -------
    cmd : str
        The command to run in docker.
    """
    container_cmd = []

    container_cmd.append(
        "export CIVIS_JOB_DATA="
        "/data/civis_job_data_${CIVIS_JOB_ID}_${CIVIS_RUN_ID}")
    container_cmd.append("mkdir -p ${CIVIS_JOB_DATA}")

    for fname, fileid in fileids.items():
        pth = os.path.dirname(fname)
        if len(pth) > 0:
            container_cmd.append("mkdir -p %s" % pth)
        container_cmd.append("civis files download %s %s" % (fileid, fname))

    if shell_cmd:
        container_cmd.append(shell_cmd)

    if repo_cmd:
        container_cmd.append("cd /app && %s && cd .." % repo_cmd)

    if os.path.isfile(script):
        container_cmd.append("chmod a+rwx %s" % remote_script)

        if script.endswith('.py'):
            cmd = 'python'
        elif script.endswith('.ipynb'):
            # Set ipython to not output color so container logs are readable.
            cmd = "ipython --InteractiveShell.colors='nocolor'"
        elif script.endswith('.r') or script.endswith('.R'):
            cmd = 'Rscript'
        else:
            cmd = 'sh'

        cstr = "%s %s" % (cmd, remote_script)
    else:
        cstr = script
    container_cmd.append(" ".join([cstr] + list(cli_args)))

    container_cmd.append("echo \"Job Output:\"")
    container_cmd.append("tar -czvf ${CIVIS_JOB_DATA}.tar.gz -C /data "
                         "civis_job_data_${CIVIS_JOB_ID}_${CIVIS_RUN_ID}")
    container_cmd.append(
        'if [ "$(ls -A ${CIVIS_JOB_DATA})" ]; then \\\n'
        '  civis scripts post-containers-runs-outputs '
        '${CIVIS_JOB_ID} ${CIVIS_RUN_ID} \\\n'
        '  `civis files upload ${CIVIS_JOB_DATA}.tar.gz` File; \\\nfi')

    return " && \\\n".join(container_cmd)


def _print_job(req, container_cmd, docker_image, params):
    """Print information about a container."""
    pdict = {}
    pdict.update(params)
    pdict['docker_image_name'] = docker_image
    pdict['required_resources'] = req

    print('container script config:')
    for line in yaml.dump(pdict, default_flow_style=False).split('\n'):
        if len(line) > 0:
            print('  ' + line)

    print('container script cmd:')
    for line in container_cmd.split('\n'):
        print('  ' + line)


def _post_and_run_job(req, container_cmd, docker_image, params, client):
    """Make and run a container. Returns the ID of the container."""

    try:
        jobid = client.scripts.post_containers(
            req, container_cmd, docker_image, **params)['id']
    except civis.base.CivisAPIError as e:
        print(
            "Could not create container: %s.%s: %s" % (
                e.__module__, e.__class__.__name__, e), file=sys.stderr)
        sys.exit(-1)

    try:
        client.scripts.post_containers_runs(jobid)
    except civis.base.CivisAPIError as e:
        print(
            "Could not run container %s: %s.%s: %s" % (
                jobid, e.__module__, e.__class__.__name__, e), file=sys.stderr)
        sys.exit(-1)

    return jobid


def _record_last_jobid(jobid):
    """Save the last jobid."""
    lid_name = os.path.expandvars(os.path.expanduser(LASTSCRIPTID_LOC))
    lid_path = os.path.split(lid_name)[0]
    if not os.path.exists(lid_path):
        os.makedirs(lid_path)
    try:
        with open(lid_name, 'w') as fp:
            fp.write('%d\n' % jobid)
    except Exception:
        pass


def _sub_job(script, remote_script, cli_args, cli_params, client, dry_run):
    """Submit a container job via the CLI.

    Parameters
    ----------
    script : str
        The local path to the script or a shell command.
    remote_script : str
        The path to put the script in the container. This could differ
        from the local path for notebooks since they are reformatted
        locally before being executed.
    cli_args : list or None
        A list of command line arguments.
    cli_params : dict
        A dictionary of command line options.
    client : civis.APIClient
        An API client instance to use.
    dry_run : bool
        If True, print info about the script and do not submit it.

    Returns
    -------
    jobid : int or None
        The ID of the script. None is returned if no script is submitted.
    """

    # Set cli_args.
    cli_args = cli_args or []

    # Get parameters.
    if os.path.isfile(script):
        with open(script, 'r') as fp:
            params = _parse_options_from_file(fp)
    else:
        params = {}

    if script.endswith('.r') or script.endswith('.R'):
        script_type = 'r'
    else:
        script_type = 'python'

    params = _process_options(params, script_type, cli_params, client)

    # Upload the files.
    otherfiles = params.pop('files', [])
    for arg in cli_args:
        if os.path.isfile(arg):
            otherfiles.append(arg)
    if os.path.isfile(script):
        otherfiles.append(script)

    use_cache = params.pop('use_file_cache')
    if not dry_run:
        fileids = _upload_files(
            otherfiles,
            use_cache)
    else:
        # some dummy IDs
        fileids = dict(zip(otherfiles, range(len(otherfiles), -1, -1)))

    # Make sure to handle the path correctly for reformatted things.
    if os.path.isfile(script) and remote_script != script:
        fileids[remote_script] = fileids[script]
        del fileids[script]

    # Now create container job submission string.
    repo_cmd = params.pop('repo_cmd', None)
    shell_cmd = params.pop('shell_cmd', None)
    container_cmd = _build_container_cmd(
        script,
        remote_script,
        cli_args,
        fileids,
        repo_cmd if 'repo_http_uri' in params else None,
        shell_cmd)

    # Final steps.
    req = params.pop('required_resources')
    docker_image = params.pop('docker_image_name')
    if dry_run:
        _print_job(req, container_cmd, docker_image, params)
        jobid = None
    else:
        jobid = _post_and_run_job(
            req, container_cmd, docker_image, params, client)

    return jobid


def _sub(script, cli_args, cli_params):
    """Driver for the CLI submit command"""
    client = civis.APIClient(resources='all')
    dry_run = cli_params.pop('dry_run')

    if os.path.isfile(script) and script.endswith('.ipynb'):
        # Make notebooks python scripts in temp dir
        with TemporaryDirectory() as tmpdir:
            new_script = os.path.join(tmpdir, os.path.basename(script))

            with open(new_script, 'w') as fp:
                fp.write(_reformat_notebook(script))

            jobid = _sub_job(
                new_script,
                script,
                cli_args,
                cli_params,
                client,
                dry_run)
    else:
        jobid = _sub_job(
            script, script, cli_args, cli_params, client, dry_run)

    if jobid:
        print(jobid)

        # If we get here, stuff the script ID in the user's home
        # area for later.
        _record_last_jobid(jobid)
